import logging
import re
from collections import defaultdict, namedtuple
from datetime import datetime, timedelta
from typing import Dict, List, Tuple

from latimes.config import LatimesConfiguration, LatimesOutputFormatting
from latimes.exceptions import InvalidTimeStringException

TIEMPO_REGEX = re.compile(
    r"^((?P<dia>[a-zA-Z]+)|(?P<fecha>\d{1,2})\sde\s(?P<mes>[a-zA-Z]+))\s(?P<hora>[0-9]{1,2})(?::(?P<minutes>[0-9]{1,2}))?\s(?P<ampm>(am|pm|AM|PM))$"
)
DIAS = {
    dia: valor
    for valor, dia in enumerate(
        ["lunes", "martes", "miercoles", "jueves", "viernes", "sabado", "domingo"]
    )
}
MESES = {
    mes: valor + 1
    for valor, mes in enumerate(
        [
            "enero",
            "febrero",
            "marzo",
            "abril",
            "mayo",
            "junio",
            "julio",
            "agosto",
            "setiembre",
            "octubre",
            "noviembre",
            "diciembre",
        ]
    )
}
DIA_DOMINGO = 6


def convert_times(cadena_tiempo: str, configuration: LatimesConfiguration) -> str:
    try:
        tiempo_usuario = interpreta_cadena_tiempo(cadena_tiempo)
    except ValueError as value_error:
        raise InvalidTimeStringException() from value_error
    tiempos = transforma_zonas_horarias(tiempo_usuario, configuration)
    return format_results(tiempo_usuario, tiempos, configuration.output_formatting)


def interpreta_cadena_tiempo(cadena_tiempo: str) -> datetime:
    match = TIEMPO_REGEX.match(cadena_tiempo)
    today = datetime.today()

    logging.info(f"Today's date is {today.isoformat()}")

    if not match:
        raise ValueError(f'"{cadena_tiempo}" is not a valid time string')

    valores = match.groupdict()

    if valores["dia"] is not None:
        dia_usuario = DIAS[valores["dia"]]

        dia_actual = today.weekday()
        if dia_usuario > dia_actual:
            dias_faltantes = dia_usuario - dia_actual
            fecha_solicitada = today + timedelta(days=dias_faltantes)
        else:
            dias_para_domingo = DIA_DOMINGO - dia_actual + 1
            fecha_solicitada = today + timedelta(days=dias_para_domingo + dia_usuario)
    elif valores["fecha"] is not None and valores["mes"] is not None:
        mes_usuario = MESES[valores["mes"]]
        dia_usuario = int(valores["fecha"])

        fecha_solicitada = datetime(today.year, mes_usuario, dia_usuario)

    minutes = int(valores["minutes"] or 0)
    hora = int(valores["hora"]) + (0 if valores["ampm"] == "am" else 12)

    return datetime(
        fecha_solicitada.year,
        fecha_solicitada.month,
        fecha_solicitada.day,
        hora,
        minutes,
    )


def transforma_zonas_horarias(
    valor_final: datetime, configuration: LatimesConfiguration
) -> List[Tuple[str, datetime]]:
    tiempos = []
    valor_localizado = configuration.starting_timezone.localize(valor_final)
    for pais, zona_horaria in configuration.convert_to.items():
        tiempos.append((pais, valor_localizado.astimezone(zona_horaria)))

    return tiempos


DateDiff = namedtuple("DateDiff", ["time", "day_difference"])


def format_results(
    anchor_datetime: datetime,
    results: List[Tuple[str, datetime]],
    output_formatting: LatimesOutputFormatting,
) -> str:
    aggregates: Dict[DateDiff, List[str]] = defaultdict(list)

    for pais, time in results:
        restored_time = time.replace(tzinfo=None)
        difference = restored_time.date() - anchor_datetime.date()
        date_diff = DateDiff(restored_time, difference.days)
        aggregates[date_diff].append(pais)

    times = []
    for date_diff, time_zones in aggregates.items():
        time_formatted = date_diff.time.strftime(output_formatting.time_format_string)
        if date_diff.day_difference:
            time_formatted += "%+d" % date_diff.day_difference

        if output_formatting.aggregate:
            joint_timezones = output_formatting.aggregate_joiner.join(time_zones)
            times.append(f"{time_formatted} {joint_timezones}")
        else:
            for time_zone in time_zones:
                times.append(f"{time_formatted} {time_zone}")

    return output_formatting.different_time_joiner.join(times)



import collections.abc
from pathlib import Path
from typing import Dict, List

import yaml
from pytz import timezone


class LatimesOutputFormatting:
    def __init__(
        self,
        time_format_string: str,
        different_time_joiner: str,
        aggregate_joiner: str,
        aggregate: bool,
    ):
        self.aggregate = aggregate
        self.aggregate_joiner = aggregate_joiner
        self.time_format_string = time_format_string
        self.different_time_joiner = different_time_joiner

    def __eq__(self, other) -> bool:
        if not isinstance(other, LatimesOutputFormatting):
            return False
        return (
            other.time_format_string == self.time_format_string
            and other.different_time_joiner == self.different_time_joiner
            and other.aggregate == self.aggregate
            and other.aggregate_joiner == self.aggregate_joiner
        )

    @classmethod
    def from_dict(cls, dictionary: Dict):
        return cls(**dictionary)


class LatimesConfiguration:
    def __init__(
        self,
        starting_timezone: str,
        convert_to: List[str],
        output_formatting: LatimesOutputFormatting,
    ):
        self.output_formatting = output_formatting
        self.convert_to = {}

        for content in convert_to:
            label, _, tz = content.partition(":")
            self.convert_to[label] = timezone(tz)
        self.starting_timezone = timezone(starting_timezone)

    def __eq__(self, other) -> bool:
        if not isinstance(other, LatimesConfiguration):
            return False

        return (
            other.convert_to == self.convert_to
            and other.output_formatting == self.output_formatting
            and other.starting_timezone == self.starting_timezone
        )

    @classmethod
    def from_dict(cls, dictionary: Dict):
        output_format = LatimesOutputFormatting.from_dict(
            dictionary["output_formatting"]
        )
        return cls(
            starting_timezone=dictionary["starting_timezone"],
            convert_to=dictionary["convert_to"],
            output_formatting=output_format,
        )


TIME_FORMAT_STRING = "%H:%M"
AGGREGATE_JOINER = ""
AGGREGATE = True
DIFFERENT_TIME_JOINER = ", "

DEFAULT_VALUES = {
    "starting_timezone": "America/Mexico_City",
    "convert_to": [
        "🇲🇽:America/Mexico_City",
        "🇨🇴:America/Bogota",
        "🇨🇱:America/Santiago",
        "🇪🇨:America/Guayaquil",
        "🇵🇪:America/Lima",
        "🇦🇷:America/Argentina/Buenos_Aires",
        "🇬🇶:Africa/Malabo",
        "🇨🇷:America/Costa_Rica",
    ],
    "output_formatting": {
        "time_format_string": TIME_FORMAT_STRING,
        "aggregate_joiner": AGGREGATE_JOINER,
        "aggregate": AGGREGATE,
        "different_time_joiner": DIFFERENT_TIME_JOINER,
    },
}


def _update(anchor, updated):
    for k, v in updated.items():
        if isinstance(v, collections.abc.Mapping):
            anchor[k] = _update(anchor.get(k, {}), v)
        elif k not in anchor:
            anchor[k] = v
    return anchor


def load_config(file: Path) -> LatimesConfiguration:
    if not file or not file.exists():
        return LatimesConfiguration.from_dict(DEFAULT_VALUES)

    with open(file) as readable:
        configuration = yaml.safe_load(readable)

    output_formatting = _update(
        configuration.get("output_formatting", dict()),
        DEFAULT_VALUES["output_formatting"],
    )
    configuration["output_formatting"] = output_formatting
    return LatimesConfiguration.from_dict(configuration)


def write_config(file: Path):
    with open(file, "w", encoding="utf8") as writable:
        writable.write("# The timezones must be expressed in TZ timezone\n")
        writable.write(
            "# https://en.wikipedia.org/wiki/List_of_tz_database_time_zones\n"
        )
        yaml.safe_dump(DEFAULT_VALUES, writable)



import logging
from pathlib import Path
from typing import List

import click

from latimes.config import load_config, write_config
from latimes.exceptions import InvalidTimeStringException
from latimes.latimes import convert_times


@click.command()
@click.argument("time_string", nargs=-1, type=click.STRING)
@click.option("--config", default=None, type=click.Path(dir_okay=False))
@click.option("--create-config/--no-create-config", default=False)
@click.option("-v", "--verbose", count=True)
def main(time_string: List[str], config: str, create_config: bool, verbose: int):
    """
    TIME_STRING Este es tu tiempo en lenguaje natural
    """
    setup_logging(verbose)

    if config:
        config_file = Path(config)
    else:
        logging.info("Will try to use default configuration file config.yml")
        config_file = Path("config.yml")

    if create_config:
        write_config(config_file)
        click.echo(f"Config file generated at {str(config_file)}")
        return

    time_string = " ".join(time_string)

    try:
        configuration = load_config(config_file)
    except KeyError as keyError:
        missing_key = keyError.args[0]
        logging.error(f"Missing key {missing_key} in config file")
        raise click.Abort()

    try:
        result = convert_times(time_string, configuration)
    except InvalidTimeStringException:
        logging.error(f'"{time_string}" is not a valid time string')
        raise click.Abort()

    print(result)


def setup_logging(verbose):
    if verbose > 1:
        logging.basicConfig(level=logging.DEBUG)
    if verbose > 0:
        logging.basicConfig(level=logging.INFO)
    logging.debug("Set logging level to " + str(logging.root.level))


if __name__ == "__main__":
    main()
